from collections import namedtuple
from time import sleep
try:from typing import Union,Optional,Tuple
except ImportError:pass
from busio import UART
from digitalio import Direction,DigitalInOut
from microcontroller import Pin
class Modes:MODE_NORMAL=0,0;MODE_WAKE_UP=1,0;MODE_POWER_SAVE=0,1;MODE_SLEEP=1,1
class SerialParity:PARITY_NONE=0;PARITY_ODD=1;PARITY_EVEN=2;PARITY_DEFAULT=PARITY_NONE
class SerialBaudRate:BAUD_1200=0,1200;BAUD_2400=1,2400;BAUD_4800=2,4800;BAUD_9600=3,9600;BAUD_19200=4,19200;BAUD_38400=5,38400;BAUD_57600=6,57600;BAUD_115200=7,115200;BAUD_DEFAULT=BAUD_9600
class AirDataRate:RATE_0_3K=0;RATE_1_2K=1;RATE_2_4K=2;RATE_4_8K=3;RATE_9_6K=4;RATE_19_2K=5;RATE_DEFAULT=RATE_2_4K
class TransmissionMode:TRANSMISSION_TRANSPARENT=0;TRANSMISSION_FIXED=1;TRANSMISSION_DEFAULT=TRANSMISSION_TRANSPARENT
class WakeUpTime:WAKE_TIME_250MS=0;WAKE_TIME_500MS=1;WAKE_TIME_750MS=2;WAKE_TIME_1000MS=3;WAKE_TIME_1250MS=4;WAKE_TIME_1500MS=5;WAKE_TIME_1750MS=6;WAKE_TIME_2000MS=7;WAKE_TIME_DEFAULT=WAKE_TIME_250MS
class IODriveMode:DRIVE_OPEN=0;DRIVE_PULL=1;DRIVE_DEFAULT=DRIVE_PULL
class ForwardErrorCorrection:FEC_DISABLED=False;FEC_ENABLE=True;FEC_DEFAULT=FEC_ENABLE
CHANNEL_MIN=0
CHANNEL_MAX=31
TX_POWER_MIN=0
TX_POWER_MAX=3
E32DeviceVersion=namedtuple('E32DeviceVersion',('model','version','features'))
E32DeviceConfig=namedtuple('E32DeviceConfig',('address','uart_parity','uart_rate','data_rate','channel','tx_mode','io_drive_mode','wake_up_time','forward_error_correction','tx_power'))
class E32Device:
	_m0:0;_m1:0;_aux:0;_uart:0;_uart_pin_tx:0;_uart_pin_rx:0;_uart_buffer_size:0;_address:0;_uart_parity:0;_uart_rate:0;_data_rate:0;_channel:0;_tx_mode:0;_io_drive_mode:0;_wake_up_time:0;_fec:0;_tx_power:0
	def __init__(A,pin_m0,pin_m1,pin_aux,pin_tx,pin_rx,uart_buffer_size=64,address=0,uart_parity=SerialParity.PARITY_NONE,uart_rate=SerialBaudRate.BAUD_DEFAULT,data_rate=AirDataRate.RATE_DEFAULT,channel=0,tx_mode=TransmissionMode.TRANSMISSION_DEFAULT,io_drive_mode=IODriveMode.DRIVE_DEFAULT,wake_up_time=WakeUpTime.WAKE_TIME_DEFAULT,forward_error_correction=True,tx_power=3):
		B=pin_aux;A._m0=DigitalInOut(pin_m0);A._m1=DigitalInOut(pin_m1);A._aux=None if B is None else DigitalInOut(B);A._uart_pin_tx=pin_rx;A._uart_pin_rx=pin_tx;A._uart_buffer_size=uart_buffer_size;A._uart=None;A.prepare_uart(9600,None);A._m0.direction=Direction.OUTPUT;A._m1.direction=Direction.OUTPUT
		if A._aux is not None:A._aux.direction=Direction.INPUT
		A._address=address;A._uart_parity=uart_parity;A._uart_rate=uart_rate;A._data_rate=data_rate;A._channel=channel;A._tx_mode=tx_mode;A._io_drive_mode=io_drive_mode;A._wake_up_time=wake_up_time;A._fec=forward_error_correction;A._tx_power=tx_power;A.mode=Modes.MODE_SLEEP;A.get_version();A.update_config()
	def __del__(A):A.deinit()
	def deinit(A):
		A.reset();A._m0.deinit();A._m1.deinit()
		if A._aux is not None:A._aux.deinit()
		A._uart.deinit()
	def prepare_uart(A,baudrate,parity):
		if A._uart is not None:A.flush_uart();A._uart.deinit()
		A._uart=UART(A._uart_pin_tx,A._uart_pin_rx,parity=parity,baudrate=baudrate,receiver_buffer_size=A._uart_buffer_size)
	def flush_uart(A):A._uart.reset_input_buffer()
	def wait_aux(A,max_wait_ms=150):
		B=max_wait_ms
		if A._aux is None:sleep(B/1000);return
		C=0
		while not A._aux.value and C<B/10-1:sleep(.01);C+=1
		sleep(.01)
	def send(B,message,max_packet_size=None,wait_aux=True):
		C=message;A=max_packet_size
		if A is None:A=len(C)
		D=B._uart.write(C[:A])
		if wait_aux:B.wait_aux()
		return D
	def read(A,byte_count=-1):B=byte_count;return A._uart.read(B if B>=0 else A.in_buffer)
	@property
	def in_buffer(self):return self._uart.in_waiting
	def update_config(A,make_permanent=False,verify=True):
		C=A.mode
		if C!=Modes.MODE_SLEEP:A.mode=Modes.MODE_SLEEP
		B=bytearray(6);B[0]=192 if make_permanent else 194;B[1]=(A._address&65280)>>8;B[2]=A._address&255;B[3]=(A._uart_parity&3)<<6|(A._uart_rate[0]&7)<<3|A._data_rate&7;B[4]=A._channel&31;B[5]=(A._tx_mode&1)<<7|(A._io_drive_mode&1)<<6|(A._wake_up_time&7)<<3|(A._fec&1)<<2|A._tx_power&3;A.flush_uart()
		if A._uart.write(B)!=6:raise RuntimeError('Failed to send the parameters configuration command !')
		A.wait_aux();A.flush_uart()
		if verify:
			D=A.get_raw_config()
			if D[1:]!=B[1:]:raise RuntimeError("The module doesn't return the new config !")
		if C!=Modes.MODE_SLEEP:A.mode=C
	def get_raw_config(A):
		B=A.mode
		if B!=Modes.MODE_SLEEP:A.mode=Modes.MODE_SLEEP
		A._uart.write(b'\xc1\xc1\xc1');A.wait_aux()
		if A._uart.in_waiting!=6:raise RuntimeError(f"The operating parameters request returned {A._uart.in_waiting} byte(s) instead of 6 !")
		C=A._uart.read(6)
		if B!=Modes.MODE_SLEEP:A.mode=B
		return C
	def get_config(B):A=B.get_raw_config();return E32DeviceConfig(A[1:3],(A[3]&192)>>6,(A[3]&56)>>3,A[3]&7,A[4]&31,(A[5]&128)>>7,(A[5]&64)>>6,(A[5]&56)>>3,(A[5]&4)>>2,A[5]&3)
	def get_raw_version(A):
		B=A.mode
		if B!=Modes.MODE_SLEEP:A.mode=Modes.MODE_SLEEP
		A._uart.write(b'\xc3\xc3\xc3');A.wait_aux()
		if A._uart.in_waiting!=4:raise RuntimeError(f"The operating parameters request returned {A._uart.in_waiting} byte(s) instead of 4 !")
		C=A._uart.read(4)
		if B!=Modes.MODE_SLEEP:A.mode=B
		return C
	def get_version(B):
		A=B.get_raw_version()
		if A[0]!=195:raise RuntimeError('The version data has a bad response code !')
		if A[1]!=50:raise RuntimeError('The version data has an invalid model !')
		return E32DeviceVersion(A[1],A[2],A[3])
	def reset(A,stay_in_sleep_mode=True):
		B=A.mode
		if B!=Modes.MODE_SLEEP:A.mode=Modes.MODE_SLEEP
		A._uart.write(b'\xc4\xc4\xc4');A.wait_aux(1500)
		if not stay_in_sleep_mode and B!=Modes.MODE_SLEEP:A.mode=B
	@property
	def mode(self):return self._m0.value,self._m1.value
	@mode.setter
	def mode(self,value):B=value;A=self;A._m0.value=B[0];A._m1.value=B[1];A.wait_aux();A.prepare_uart(baudrate=9600 if A.mode==Modes.MODE_SLEEP else A._uart_rate[1],parity=None if A.mode==Modes.MODE_SLEEP or A.uart_parity==SerialParity.PARITY_NONE else UART.Parity.ODD if A.uart_parity==SerialParity.PARITY_ODD else UART.Parity.EVEN);A.flush_uart()
	@property
	def address(self):return self._address
	@address.setter
	def address(self,value):
		A=value
		if not 0<=A<=65535:raise ValueError(f"The 'address' isn't between 0x0000 and 0xFFFF !  (It's {A})")
		self._address=A;self.update_config()
	@property
	def uart_parity(self):return self._uart_parity
	@uart_parity.setter
	def uart_parity(self,value):
		A=value
		if not 0<=A<=3:raise ValueError(f"The 'uart_parity' isn't between 0 and 3 !  (It's {A})")
		self._uart_parity=A;self.update_config()
	@property
	def uart_rate(self):return self._uart_rate
	@uart_rate.setter
	def uart_rate(self,value):
		A=value
		if not 0<=A[0]<=7:raise ValueError(f"The 'uart_rate' isn't between 0 and 7 !  (It's {A[0]})")
		self._uart_rate=A;self.update_config()
	@property
	def data_rate(self):return self._data_rate
	@data_rate.setter
	def data_rate(self,value):
		A=value
		if not 0<=A<=7:raise ValueError(f"The 'data_rate' isn't between 0 and 7 !  (It's {A})")
		self._data_rate=A;self.update_config()
	@property
	def channel(self):return self._channel
	@channel.setter
	def channel(self,value):
		A=value
		if not CHANNEL_MIN<=A<=CHANNEL_MAX:raise ValueError(f"The 'channel' isn't between {CHANNEL_MIN} and {CHANNEL_MAX} !  (It's {A})")
		self._channel=A;self.update_config()
	@property
	def tx_mode(self):return self._tx_mode
	@tx_mode.setter
	def tx_mode(self,value):
		A=value
		if not 0<=A<=1:raise ValueError(f"The 'tx_mode' isn't between 0 and 1 !  (It's {A})")
		self._tx_mode=A;self.update_config()
	@property
	def io_drive_mode(self):return self._io_drive_mode
	@io_drive_mode.setter
	def io_drive_mode(self,value):
		A=value
		if not 0<=A<=1:raise ValueError(f"The 'io_drive_mode' isn't between 0 and 1 !  (It's {A})")
		self._io_drive_mode=A;self.update_config()
	@property
	def wake_up_time(self):return self._wake_up_time
	@wake_up_time.setter
	def wake_up_time(self,value):
		A=value
		if not 0<=A<=7:raise ValueError(f"The 'wake_up_time' isn't between 0 and 7 !  (It's {A})")
		self._wake_up_time=A;self.update_config()
	@property
	def forward_error_correction(self):return self._fec
	@forward_error_correction.setter
	def forward_error_correction(self,value):self._fec=value;self.update_config()
	@property
	def tx_power(self):return self._tx_power
	@tx_power.setter
	def tx_power(self,value):
		A=value
		if not 0<=A<=3:raise ValueError(f"The 'tx_power' isn't between 0 and 3 !  (It's {A})")
		self._tx_power=A;self.update_config()